using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Globalization;

namespace RTadeusiewicz.NN.NeuralNetworks
{
    /// <summary>
    /// Ta klasa reprezentuje zbir uczcy, uywany do uczenia sieci.
    /// </summary>
    /// <remarks>
    /// Klasa ta jest po prostu list elementw uczcych, wzbogacon o kilka
    /// uytecznych metod pomocniczych.
    /// 
    /// Zbir uczcy moe zosta wczytany z pliku. Plik ten musi mie nastpujcy
    /// format:
    /// 
    /// <code>
    /// 3, 2
    /// pierwszy obiekt
    /// 3, 4.5, 6.3
    /// 0.5, 0.23
    /// drugi obiekt
    /// 3, 2, 1
    /// 0, 0
    /// </code>
    /// 
    /// Pierwszy wiersz pliku zawiera kolejno ilo wej i wyj sieci neuronowej,
    /// dla ktrej przeznaczony jest dany zbir uczcy. Kolejne linie opisuj
    /// poszczeglne obiekty cigu uczcego. Opis kadego obiektu skada si z:
    /// 
    /// <list type="bullet">
    /// <item><description>komentarza (nazwy obiektu)</description></item>
    /// <item><description>wartoci wej</description></item>
    /// <item><description>prawidowych wartoci wyj</description></item>
    /// </list>
    /// 
    /// Ilo wartoci wej i wyj dla kadego obiektu musi by zgodna z wymiarami
    /// sieci podanymi w pierwszej linii.
    /// </remarks>
    public class TeachingSet : List<TeachingSet.Element>
    {
        /// <summary>
        /// Element cigu uczcego. Zawiera wejcia, prawidowe wyjcie i komentarz
        /// (nazw obiektu opisywanego przez wejcia).
        /// </summary>
        public struct Element
        {
            /// <summary>
            /// Wartoci sygnaw, jakie powinny zosta podane na wejcia sieci
            /// neuronowej. Mog wymaga normalizacji.
            /// </summary>
            public double[] Inputs;

            /// <summary>
            /// Wartoci, jakich oczekujemy na wyjciu sieci neuronowej.
            /// </summary>
            public double[] ExpectedOutputs;

            /// <summary>
            /// Komentarz. Moe by to nazwa obiektu reprezentowanego przez ten
            /// element.
            /// </summary>
            public string Comment;
            
            public Element Clone()
            {
                Element ret = this;
                ret.Inputs = (double[])Inputs.Clone();
                ret.ExpectedOutputs = (double[])ExpectedOutputs.Clone();
                return ret;
            }
        }

        /// <summary>
        /// Tworzy nowy zbir uczcy.
        /// </summary>
        /// <param name="inputCount">Ilo wej sieci neuronowej, dla ktrej
        /// przeznaczony jest ten zbir.</param>
        /// <param name="outputCount">Ilo wyj sieci neuronowej, dla ktrej
        /// przeznaczony jest ten zbir.</param>
        public TeachingSet(int inputCount, int outputCount)
        {
            _inputCount = inputCount;
            _outputCount = outputCount;
        }

        public TeachingSet(TeachingSet set): base(set.Count)
        {
            _inputCount = set._inputCount;
            _outputCount = set._outputCount;
            foreach (Element elem in set)
                Add(elem.Clone());
        }

        // Przechowuje warto wasnoci InputCount
        private int _inputCount;

        /// <summary>
        /// Ilo wej sieci neuronowej, dla ktrej przeznaczony jest ten zbir.
        /// Wasno ta ma jedynie charakter informacyjny i wcale nie musi
        /// odzwierciedla stanu faktycznego, sczeglnie jeli zostaa zmieniona.
        /// Istnieje jedynie gwarancja, e po odczytaniu zbioru z pliku wasno
        /// ta bdzie miaa prawidow warto.
        /// </summary>
        public int InputCount
        {
            get { return _inputCount; }
            set { _inputCount = value; }
        }

        // Przechowuje warto wasnoci OutputCount.
        private int _outputCount;

        /// <summary>
        /// Ilo wyj sieci neuronowej, dla ktrej przeznaczony jest ten zbir.
        /// Wasno ta ma jedynie charakter informacyjny i wcale nie musi
        /// odzwierciedla stanu faktycznego, sczeglnie jeli zostaa zmieniona.
        /// Istnieje jedynie gwarancja, e po odczytaniu zbioru z pliku wasno
        /// ta bdzie miaa prawidow warto.
        /// </summary>
        public int OutputCount
        {
            get { return _outputCount; }
            set { _outputCount = value; }
        }

        // Przetwarza pojedyncz lini z pliku opisujcego zbir uczcy. Zwraca
        // tablic liczb zawartych w tej linii lub wyrzuca wyjtek o podanej treci,
        // jeli ich ilo nie zgadza si z oczekiwaniami.
        private static double[] ParseLine(TextReader reader, int expectedLength,
            string exceptionString)
        {
            /* Musimy to zrobi, aby format pliku by uniwersalny (rne narody
             * w rny sposb zapisuj liczby dziesitne). */
            NumberFormatInfo invariantFormat =
                CultureInfo.InvariantCulture.NumberFormat;

            String line = reader.ReadLine();
            if (line == null)
                return null;
            if (line == "")
                throw new ApplicationException(
                    "Empty line encountered where the numbers were expected.");

            /* Rozbijamy lini na liczby rozdzielone przecinkami. Spacje nam nie
             * przeszkadzaj. */
            string[] fields = line.Split(',');
            if (fields.Length != expectedLength)
                throw new ApplicationException(exceptionString);
            double[] ret = new double[expectedLength];
            for (int i = 0; i < fields.Length; i++)
                ret[i] = double.Parse(fields[i], invariantFormat);
            return ret;
        }

        /// <summary>
        /// Wczytuje zbir uczcy z pliku.
        /// </summary>
        /// <param name="fileName">Nazwa pliku.</param>
        /// <exception cref="ApplicationException">Plik ma nieprawidowy format.
        /// </exception>
        /// <returns>Wczytany zbir uczcy lub <c>null</c>, jeli plik jest pusty.
        /// </returns>
        public static TeachingSet FromFile(string fileName)
        {
            using (StreamReader sr = new StreamReader(fileName))
            {
                return TeachingSet.FromText(sr);
            }
        }

        /// <summary>
        /// Wczytuje zbir uczcy ze strumienia.
        /// </summary>
        /// <param name="stream">Strumie zawierajcy zbir uczcy</param>
        /// <exception cref="ApplicationException">Strumie ma nieprawidowy format.
        /// </exception>
        /// <returns>Wczytany zbir uczcy lub <c>null</c>, jeli strumie jest pusty.
        /// </returns>
        public static TeachingSet FromStream(Stream stream)
        {
            using (StreamReader sr = new StreamReader(stream))
            {
                return TeachingSet.FromText(sr);
            }
        }

        /// <summary>
        /// Interpretuje podany tekst jako opis zbioru uczcego.
        /// </summary>
        /// <param name="reader">Obiekt odczytujcy tekst.</param>
        /// <exception cref="ApplicationException">Tekst ma nieprawidowy format.
        /// </exception>
        /// <returns>Wczytany zbir uczcy lub <c>null</c>, jeli tekst jest pusty.
        /// </returns>
        public static TeachingSet FromText(TextReader reader)
        {
            try
            {
                // Pierwsza linia - wymiary sieci.
                string line = reader.ReadLine();
                if (line == null)
                    return null;

                string[] fields = line.Split(',');
                if (fields.Length != 2)
                    throw new ApplicationException("First line must contain two numbers.");
                int inputCount = int.Parse(fields[0]);
                int outputCount = int.Parse(fields[1]);

                TeachingSet ret = new TeachingSet(inputCount, outputCount);

                // Poszczeglne obiekty.
                while ((line = reader.ReadLine()) != null)
                {
                    Element elem;
                    elem.Comment = line;

                    elem.Inputs =
                        ParseLine(reader, inputCount, "Invalid number of inputs.");
                    if (elem.Inputs == null)
                        break;

                    elem.ExpectedOutputs =
                        ParseLine(reader, outputCount, "Invalid number of outputs.");
                    if (elem.ExpectedOutputs == null)
                        break;

                    ret.Add(elem);
                }
                return ret;
            }
            catch (ApplicationException) { throw; }
            catch (Exception ex)
            {
                throw new ApplicationException("The file is corrupted.", ex);
            }
        }

        public void Normalize()
        {
            foreach (Element elem in this)
                Neuron.Normalize(elem.Inputs);
        }
    }
}
